/*
 * this file is a part of yabbs - yet another bulletin board system.
 * Copyright (C) 1993 Alex Wetmore.  
 * email: aw2t@andrew.cmu.edu
 * address: 6 rech ave
 *          oreland pa 19075
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * standard library for yabbs clients.  all functions designed for use by
 * clients start with bbs_ and are defined in yabbslib.h.  All internal
 * function start with bbsi_ or c_.  All functions expected to be provided
 * by the client start with cli_.
 */

#include <stdio.h>                          /* standard i/o stuff           */
#include <unistd.h>                         /* more standard stuff          */
#include <stdlib.h>                         /* standard library             */
#include <ctype.h>                          /* character type macros        */
#include <strings.h>                        /* string utils                 */
#include <time.h>                           /* time stuff                   */
#include <sys/param.h>                      /* htons()                      */
#include <sys/socket.h>                     /* sockets                      */
#include <netinet/in.h>                     /* tcp/ip stuff                 */
#include <netdb.h>                          /* more net stuff               */
#include <stdarg.h>                         /* variable argument macros     */
#include "network.h"                        /* packet types for server      */

#include "yabbslib.h"                       /* bbs header file              */

#define YABBS
#ifndef TRUE
#define TRUE 1                              /* you might have these         */
#define FALSE 0
#endif

int     bbs_s;                              /* server socket                */
char    bbs_waitkey[64];                    /* message to wait for          */
int     bbs_msgnum, bbs_lastmsgnum;         /* set by getmsghdrs            */
int     bbs_loggedout = FALSE;              /* set when we have logged out  */
int     bbs_key = 0;                        /* message key                  */
int     bbs_debug = 0;                      /* don't print debug messages   */

/*
 * function: char bbsi_upper(char i);
 * input:    any valid character
 * returns:  lowercase version of character
 * notes:    internal function, should only be used by yabbslib functions
 */

char bbsi_upper(char i) {
    if ((i >= 'a') && (i <= 'z')) i = i + ('A' - 'a');
    return i;
}

/*
 * function: char *bbsi_itoa(int i);
 * input:    any integer
 * output:   string representing integer
 * notes:    internal function, should only be used by yabbslib functions
 *           this function returns a pointer to static buffer
 */
char *bbsi_itoa(int i) {
    static char buf[255];

    sprintf(buf, "%i", i);
    return buf;
}

/*
 * function: char *bbsi_itoa(int i);
 * input:    any unsigned number
 * output:   string representing unsigned number
 * notes:    internal function, should only be used by yabbslib functions
 *           this function returns a pointer to static buffer
 */
char *bbsi_utoa(unsigned long i) {
    static char buf[255];

    sprintf(buf, "%u", i);
    return buf;
}

/*
 * function: char* bbsi_nextkey();
 * input:    none
 * output:   the next unique key for use in server messages
 * notes:    internal function, should only be used by yabbslib functions
 *           this function returns a pointer to static buffer
 */
char *bbsi_nextkey(void) {
    static char key[255];

    sprintf(key, "%i", ++bbs_key);
    return key;
}

/*
 * function: char* bbsi_nextkey();
 * input:    none
 * output:   the current key for use in server messages
 * notes:    internal function, should only be used by yabbslib functions
 *           this function returns a pointer to static buffer
 */
char *bbsi_thiskey(void) {
    static char key[255];

    sprintf(key, "%i", bbs_key);
    return key;
}

/*
 * function: bbsi_csendp(int n, char *key, char *command, char *arg, ...);
 * input:    n is the number of arguments (including the command)
 *           key a key used by the server when responding to this message
 *           command is the command to do
 *           args are a variable number of arguments for the command
 *
 * notes:    for more information on this see the protocol documentation if
 *           they're ever written.
 */
void bbsi_csendp(int n, char *key, char *command, ...) {
    va_list ap;
    char *parameter, p2[1024], buf[1024];
    int i, x, y;

    va_start(ap, command);
    strcpy(buf, key); strcat(buf, ":");
    strcat(buf, command);

    for (i = 1; i < n; i++) {
        strcat(buf, ":"); 
        parameter = va_arg(ap, char *);
        for (x = 0, y = 0; parameter[x] != 0; x++) {
            if (parameter[x] == ':') p2[y++] = '\\'; 
            if (parameter[x] == '\\') p2[y++] = '\\'; 
            p2[y++] = parameter[x];
        }
        p2[y] = 0;
        strcat(buf, p2);
    }
    strcat(buf, "\n");
    if (bbs_debug & DBG_SEND) printf("send: %s", buf);
    send(bbs_s, buf, strlen(buf), 0);
}

/*
 * function: char *bbsi_creadp()
 * returns:  a pointer to a string from the server
 * notes:    the string is statically allocated
 */
char *bbsi_creadp() {
    int reading = 1, i = 0;
    char x, c;
    static char buf[1024];

    /* read characters until we get a NL or end of a packet                 */
    while (reading) {
        x = read(bbs_s, &c, 1);
        if ((x < 1) || (c == '\n') || (i == 510)) reading = 0;
        else buf[i++] = c;
    }
    buf[i] = 0;

    /* bleed off any extra characters before the packet end                 */
    if ((c != '\n') && (i == 510))
        do read(bbs_s, &c, 1); while (c != '\n');

    /* if we read 0 characters then the server died                         */
    if (i == 0) {
        if (!bbs_loggedout) 
            cli_fatalerror("The yabbs server closed your connection.");
        else cli_fatalerror("connection closed");
    }

    if (bbs_debug & DBG_RECV) printf("recv: %s\n", buf);
    return buf;
}

/*
 * -----------------------------------------------------------------------
 * all of the functions for dealing with protocol commands from the server
 * -----------------------------------------------------------------------
 */

int c_yabbshi(char **argv) {
    return 1;
}

int c_bad(char **argv) {
    char buf[255];

    sprintf(buf, "SERVER WARNING: %s", argv[1]);
    cli_error(buf);
}

int c_talk(char **argv) {
    char buf[255];

    if (strcmp(argv[1], "PUBLIC") == 0) {
        cli_talkline(argv[2], argv[3], argv[4], TL_PUBLIC);
    } else if (strcmp(argv[1], "PRIVATE") == 0) {
        cli_talkline(argv[2], argv[3], argv[4], TL_PRIVATE);
    } else if (strcmp(argv[1], "ACTION") == 0) {
        cli_talkline(argv[2], argv[3], argv[4], TL_ACTION);
    } else if (strcmp(argv[1], "HIDE") == 0) {
        cli_talkline(argv[2], argv[3], argv[4], TL_HIDE);
    } else if (strcmp(argv[1], "SHOW") == 0) {
        cli_talkline(argv[2], argv[3], argv[4], TL_SHOW);
    } else if (strcmp(argv[1], "JOIN") == 0) {
        cli_talkline(argv[2], argv[3], argv[4], TL_JOIN);
    } else if (strcmp(argv[1], "LEAVE") == 0) {
        cli_talkline(argv[2], argv[3], argv[4], TL_LEAVE);
    } else {
        sprintf(buf, "*** unknown message type TALK:%s from server", argv[1]);
        cli_error(buf);
    }
}

int c_page(char **argv) {
    cli_page(argv[1]);
}

int c_who(char **argv) {
    char buf[255], atime[32];
    int idletime;
    time_t tp;

    if (strcmp(argv[1], "START") == 0) {
        strcpy(bbs_waitkey, bbsi_thiskey());
    } else if (strcmp(argv[1], "LINE") == 0) {
        tp = (time_t) atoi(argv[4]);
        strcpy(atime, (char *) ctime(&tp));
        atime[strlen(atime) - 1] = 0;
        idletime = atoi(argv[5]);
        cli_wholine(argv[2], argv[3], atime, idletime);
        strcpy(bbs_waitkey, bbsi_thiskey());
    } else if (strcmp(argv[1], "END") == 0) {
    } else {
        sprintf(buf, "yabbslib: c_who: invalid packet type WHO:%s from server", 
            argv[1]);
        cli_error(buf);
    }
}

int c_channels(char **argv) {
    char buf[255];

    if (strcmp(argv[1], "START") == 0) {
        strcpy(bbs_waitkey, bbsi_thiskey());
    } else if (strcmp(argv[1], "LINE") == 0) {
        cli_channelline(argv[2], argv[3]);
        strcpy(bbs_waitkey, bbsi_thiskey());
    } else if (strcmp(argv[1], "END") == 0) {
    } else {
        sprintf(buf, 
            "yabbslib: c_channels: invalid packet type CHANNELS:%s from server", 
            argv[1]);
        cli_error(buf);
    }
}

int c_msg(char **argv) {
    char buf[255];

    if (strcmp(argv[1], "START") == 0) {
    } else if (strcmp(argv[1], "TO") == 0) {
        cli_msgline(argv[2], ML_TO);
    } else if (strcmp(argv[1], "FROM") == 0) {
        cli_msgline(argv[2], ML_FROM);
    } else if (strcmp(argv[1], "TITLE") == 0) {
        cli_msgline(argv[2], ML_TITLE);
    } else if (strcmp(argv[1], "TIME") == 0) {
        cli_msgline(argv[2], ML_TIME);
    } else if (strcmp(argv[1], "DELETED") == 0) {
        cli_msgline(argv[2], ML_DELETED);
    } else if (strcmp(argv[1], "LINE") == 0) {
        cli_msgline(argv[2], ML_LINE);
    } else if (strcmp(argv[1], "END") == 0) {
        return M_OK;
    } else if (strcmp(argv[1], "MSGOOB") == 0) {
        return M_OOB;
    } else if (strcmp(argv[1], "PRIVATE") == 0) {
        return M_PRIVATE;
    } else {
        sprintf(buf, "yabbslib: c_msg: invalid packet type MSG:%s from server", 
            argv[1]);
        cli_error(buf);
        return M_ERR;
    }
    /* recurse until we get our whole message */
    strcpy(bbs_waitkey, bbsi_thiskey());
}

int c_msghdr(char **argv) {
    if (bbs_msgnum != bbs_lastmsgnum) {
        strcpy(bbs_waitkey, bbsi_thiskey());
        bbs_msgnum++;
    }

    if (strcmp(argv[3], "OOB") == 0) {
        bbs_waitkey[0] = 0;
        return M_OOB;                           /* invalid message number   */
    } else if (strcmp(argv[3], "PRIVATE") == 0) {
        return M_PRIVATE;                       /* private message          */
    }

    cli_msghdrline(bbs_msgnum - 1, argv[1], argv[2], argv[3], argv[4]);
    return M_OK;
}

int c_msgnum(char **argv) {
    int msgnum;

    msgnum = atoi(argv[2]);

    return msgnum;
}

int c_delmsgret(char **argv) {
    if (argv[3][0] == 'P') return M_PRIVATE;
    else return M_OK;
}

int c_basetitle(char **argv) {
    int baseid;

    if (strcmp(argv[1], "END") == 0) {
        return 0;
    } else if (strcmp(argv[1], "START") == 0) {
        strcpy(bbs_waitkey, bbsi_thiskey());
    } else if (strcmp(argv[1], "LINE") == 0) {
        baseid = atoi(argv[2]);
        cli_msgbasename(baseid, argv[3]);
        strcpy(bbs_waitkey, bbsi_thiskey());
    }
}

int c_useropts(char **argv) {
    cli_useroptions(argv[1], atoi(argv[2]), atoi(argv[3]), argv[4]);
}

int c_login(char **argv) {
    char buf[255];

    if (strcmp(argv[1], "OK") == 0) return TRUE;
    else {
        sprintf(buf, "problem logging in: %s", argv[1]);
        cli_error(buf);
        return FALSE;
    }
}

int c_lastmsg(char **argv) {
    return atoi(argv[2]);
}

int c_newuser(char **argv) {
    char buf[255];

    if (strcmp(argv[1], "OK") == 0) return TRUE; 
    else {
        sprintf(buf, "problem creating new user: %s", argv[1]);
        cli_error(buf);
        return FALSE;
    }
}

int c_founduser(char **argv) {
    if (strcmp(argv[1], "START") == 0) {
        strcpy(bbs_waitkey, bbsi_thiskey());
    } else if (strcmp(argv[1], "END") == 0) {
        return 0;
    } else {
        cli_founduser(argv[3]);
        strcpy(bbs_waitkey, bbsi_thiskey());
    }
}

int c_userexist(char **argv) {
    if (strcmp(argv[2], "OK") == 0) return TRUE; else return FALSE;
}

int c_txtfile(char **argv) {
    if (strcmp(argv[2], "END") == 0) return;
    else if (strcmp(argv[2], "LINE") == 0) cli_textline(argv[3]);
    strcpy(bbs_waitkey, bbsi_thiskey());
}

int c_planfile(char **argv) {
    if (strcmp(argv[2], "END") == 0) return;
    else if (strcmp(argv[2], "LINE") == 0) cli_planline(argv[3]);
    else if (strcmp(argv[2], "BADNAME") == 0) {
        cli_planline("No plan file<nl>");
        return;
    }
    strcpy(bbs_waitkey, bbsi_thiskey());
}

int c_gfileindex(char **argv) {
    if (strcmp(argv[1], "START") == 0) {
    } else if (strcmp(argv[1], "NOFILE") == 0) {
        cli_error("Problem accessing gfile");
        return 0;
    } else if (strcmp(argv[1], "LINE") == 0) {
        cli_gfileindexline(argv[3], argv[4], argv[2]);
    } else if (strcmp(argv[1], "END") == 0) {
        return 1;
    }
    strcpy(bbs_waitkey, bbsi_thiskey());
}

int c_gfile(char **argv) {
    if (strcmp(argv[1], "START") == 0) {
    } else if (strcmp(argv[1], "NOFILE") == 0) {
        cli_error("Problem accessing gfile");
        return 0;
    } else if (strcmp(argv[1], "LINE") == 0) {
        cli_gfileline(argv[2]);
    } else if (strcmp(argv[1], "END") == 0) {
        return 1;
    }
    strcpy(bbs_waitkey, bbsi_thiskey());
}

int c_shutdown(char **argv) {
    char buf[255];

    sprintf(buf, "system shutdown: %s", argv[1]);
    cli_error(buf);
}

/*
 * ------------------------------------------------------------------------
 * functions that people should use to interact with the server.  many of
 * these return data though "call-back" functions, all of which start with
 * cli_
 * ------------------------------------------------------------------------
 */

/*
 * this deals with incoming messages from the server.  it should be called
 * whenever your select or SIGIO setup discovers that there is something
 * waiting on the server soocket (bbs_s).
 */
int bbs_handlemsg(void) {
    char *token[1024], **tp = token, *p, *key;
    int x, i, j, l, ci, plen, done, rcode, moremsgs;
    char fullmsg[1024], msg[1024], buf[255]; 

    moremsgs = TRUE;
    while (moremsgs) {
        strcpy(fullmsg, bbsi_creadp());

        for (i = j = 0; fullmsg[i] != 0; i++, j++) {
            if (fullmsg[i] == '\\') {
                switch (fullmsg[i + 1]) {
                    case 'n': msg[j] = '\n';            break;
                    case 't': msg[j] = '\t';            break;
                    case 'b': msg[j] = '\b';            break;
                    default : msg[j] = fullmsg[i + 1];  break;
                }
                i++;
            } else if (fullmsg[i] == ':') {
                msg[j] = 0x18;
            } else msg[j] = fullmsg[i];
        }
        msg[j] = 0;

        done = FALSE;
        plen = strlen(msg);
    
        if (strchr(msg, 0x18) == NULL) {
            cli_fatalerror("this server speaks an older version of the yabbs protocol");
        } else {
            p = msg;                        /* parse string p               */
            tp = token;
            key = strsep(&p, "\x18");       /* read the command key         */
            *(tp++) = strsep(&p, "\x18");   /* read first word              */
            for (x = 0, l = strlen(token[0]); x < l; x++) 
                token[0][x] = bbsi_upper(token[0][x]); 
            for (ci = 0; ci < C_CNUM; ci++) /* is it in our table?          */
                if (strcmp(s_ctable[ci].cname, token[0]) == 0) break;
            if (ci == C_CNUM) {             /* no command with that name    */
                sprintf(buf, "yabbslib: bbs_handlemsg: invalid packet type %s from server", 
                    token[0]);
                cli_error(buf);
                return;
            }                               /* otherwise ci points to cmd   */
            while (p != NULL) {
                (char *) *(tp++) = (char *) strsep(&p, "\x18");
            }
        }
    
        /* clear the bbs_waitkey flag if this is the right type of message      */
        if (strcmp(bbs_waitkey, key) == 0) *bbs_waitkey = 0;

        /* call the function to handle this type of message                 */
        rcode = s_ctable[ci].f(token);

        if (*bbs_waitkey == 0) moremsgs = FALSE;
    }
    return rcode;
}

/*
 * function: bbs_logout()
 * notes:    closes the connection to the yabbs server
 */
void bbs_logout() {
    bbs_loggedout = TRUE;
    bbsi_csendp(1, bbsi_nextkey(), "LOGOUT");
}

/*
 * function: int bbs_login(char *name, char *pass)
 * input:    name - account name to login under
 *           pass - password to use for account
 * output:   TRUE if the login was valid, FALSE otherwise
 * notes:    if they weren't given access an error message will be send with
 *           cli_error() callback.
 */
int bbs_login(char *name, char *pass) {
    strcpy(bbs_waitkey, bbsi_nextkey());
    bbsi_csendp(3, bbs_waitkey, "LOGIN", name, pass);
    return bbs_handlemsg();
}

/*
 * function: bbs_connect(char *server);
 * input:    server - the server to connect to
 * returns:  none
 * notes:    if this couldn't connect it will return a fatal error explaining
 *           why.
 */
void bbs_connect(char *server) {
    struct sockaddr_in sin;
    struct hostent *hp;

    hp = gethostbyname(server);
    if (hp == NULL) {
        cli_fatalerror("yabbslib: bbs_connect: invalid server hostname\n");
    }
    bzero((char *) &sin, sizeof(sin));
    bcopy(hp->h_addr, (char *) &sin.sin_addr, hp->h_length);
    sin.sin_family = hp->h_addrtype;
    sin.sin_port = htons(YABBSPORT);

    bbs_s = socket(AF_INET, SOCK_STREAM, 0);

    if (connect(bbs_s, (struct sockaddr *) &sin, sizeof(sin)) == -1) {
        cli_fatalerror("The yabbs server is down.  Try again later.");
    }

    strcpy(bbs_waitkey, "_NEWCLI");
    if (!bbs_handlemsg()) {
        cli_fatalerror("Doesn't look like a yabbs server");
    }
}

/*
 * function: bbs_getmsg(int baseid, int msgnum);
 * inputs:   baseid - the message base to get the message from
 *           msgnum - the number of the message to get
 * returns:  M_OK if it got the message, M_OOB if the baseid or message number
 *           is invalid, M_PRIVATE if the message is private, or M_ERR if there
 *           is another error.  If it returns M_ERR a reason will be sent back
 *           using the cli_error callback.
 *
 * notes:    the message is sent back using the cli_msgline(type, string) 
 *           callback.  the types and corresponding strings are:
 *             ML_TO       who the message is to
 *             ML_FROM     who the message is from
 *             ML_TITLE    the title of the message
 *             ML_TIME     when the message was sent
 *             ML_DELETED  deleted status of the message
 *             ML_LINE     the lines of the message
 *           there can be multiple ML_LINE messages, but only one of each
 *           of the other types.
 */
int bbs_getmsg(int baseid, int msgnum) {
    char baseid_st[10];

    strcpy(baseid_st, bbsi_itoa(baseid));
    strcpy(bbs_waitkey, bbsi_nextkey());
    bbsi_csendp(3, bbs_waitkey, "GETMSG", baseid_st, bbsi_itoa(msgnum));
    return bbs_handlemsg();
}

/*
 * function: bbs_getmsghdr(int baseid, int msgnum);
 * inputs:   baseid - the message base to get the message from
 *           msgnum - the message number to get
 * returns:  M_OOB if the message was out of bounds, M_PRIVATE if it private,
 *           or M_OK otherwise.
 * notes:    the data is returned using the cli_msghdrline() callback.  The
 *           prototype for this should be: cli_msghdrline(int msgnum, char *to, 
 *           char *from, char *deleted, char *subject);
 * 
 *           if you are reading multiple headers in a loop please use 
 *           bbs_getmsghdrs instead.  it is faster on the server.
 */
int bbs_getmsghdr(int baseid, int msgnum) {
    char baseid_st[10], msgnum_st[10];

    strcpy(baseid_st, bbsi_itoa(baseid));
    strcpy(msgnum_st, bbsi_itoa(msgnum));
    strcpy(bbs_waitkey, bbsi_nextkey());
    bbsi_csendp(4, bbs_waitkey, "GETMSGHDRS", baseid_st, msgnum_st, msgnum_st);
    bbs_msgnum = msgnum;
    bbs_lastmsgnum = msgnum;
    return bbs_handlemsg();
}

/*
 * function: bbs_getmsghdrs(int baseid, int msgnum1, msgnum2);
 * inputs:   baseid  - the message base to get the message from
 *           msgnum1 - the first message header to get
 *           msgnum2 - the second message header to get
 * returns:  M_OOB if the message was out of bounds, M_PRIVATE if it private,
 *           or M_OK otherwise (for the last message only).
 * notes:    the data is returned using the cli_msghdrline() callback.  There
 *           will be one call to this for each message header that the user
 *           has access to read.
 */
int bbs_getmsghdrs(int baseid, int msgnum1, int msgnum2) {
    char msgnum1_st[10], msgnum2_st[10];

    strcpy(msgnum1_st, bbsi_itoa(msgnum1));
    strcpy(msgnum2_st, bbsi_itoa(msgnum2));
    strcpy(bbs_waitkey, bbsi_nextkey());
    bbsi_csendp(4, bbs_waitkey, "GETMSGHDRS", bbsi_itoa(baseid), msgnum1_st, 
        msgnum2_st);
    bbs_msgnum = msgnum1;
    bbs_lastmsgnum = msgnum2;
    return bbs_handlemsg();
}

int bbs_getgfileindex(char *path) {
    strcpy(bbs_waitkey, bbsi_nextkey());
    bbsi_csendp(2, bbs_waitkey, "GFILEINDEX", path);
    return bbs_handlemsg();
}

int bbs_getgfile(char *path) {
    strcpy(bbs_waitkey, bbsi_nextkey());
    bbsi_csendp(2, bbs_waitkey, "GETGFILE", path);
    return bbs_handlemsg();
}

void bbs_getplanfile(char *filename) {
    strcpy(bbs_waitkey, bbsi_nextkey());
    bbsi_csendp(2, bbs_waitkey, "GETPLANFILE", filename);
    bbs_handlemsg();
}

void bbs_gettextfile(char *filename) {
    strcpy(bbs_waitkey, bbsi_nextkey());
    bbsi_csendp(2, bbs_waitkey, "GETTXTFILE", filename);
    bbs_handlemsg();
}

void bbs_who(void) {
    strcpy(bbs_waitkey, bbsi_nextkey());
    bbsi_csendp(1, bbs_waitkey, "WHO");
    bbs_handlemsg();
}

void bbs_channels(void) {
    strcpy(bbs_waitkey, bbsi_nextkey());
    bbsi_csendp(1, bbs_waitkey, "CHANNELS");
    bbs_handlemsg();
}

void bbs_talkjoin(char *channel) {
    bbsi_csendp(3, bbsi_nextkey(), "TALK", "JOIN", channel);
}

void bbs_talkleave(char *channel) {
    bbsi_csendp(3, bbsi_nextkey(), "TALK", "LEAVE", channel);
}

void bbs_talkpublic(char *channel, char *text) {
    bbsi_csendp(4, bbsi_nextkey(), "TALK", "PUBLIC", channel, text);
}

void bbs_talkprivate(char *to, char *text) {
    bbsi_csendp(4, bbsi_nextkey(), "TALK", to, "", text);
}

void bbs_talkaction(char *channel, char *text) {
    bbsi_csendp(4, bbsi_nextkey(), "TALK", "ACTION", channel, text);
}

void bbs_talkshow(char *channel) {
    bbsi_csendp(3, bbsi_nextkey(), "TALK", "SHOW", channel);
}

void bbs_talkhide(char *channel) {
    bbsi_csendp(3, bbsi_nextkey(), "TALK", "HIDE", channel);
}

void bbs_talkquit(void) {
    bbsi_csendp(3, bbsi_nextkey(), "TALK", "QUIT", "");
}

void bbs_page(char *name) {
    bbsi_csendp(2, bbsi_nextkey(), "PAGE", name);
}

int bbs_getmsgnum(int baseid) {
    strcpy(bbs_waitkey, bbsi_nextkey());
    bbsi_csendp(2, bbs_waitkey, "GETMSGNUM", bbsi_itoa(baseid));
    return bbs_handlemsg();
}

void bbs_addmsg(int baseid, char *to, char *subject, char *msg) {
    char buf[255];
    int bp, x, msglen;

    bbsi_csendp(3, bbsi_nextkey(), "ADDMSG", "START", bbsi_itoa(baseid));
    bbsi_csendp(3, bbsi_nextkey(), "ADDMSG", "TO", to);
    bbsi_csendp(3, bbsi_nextkey(), "ADDMSG", "TITLE", subject);

    msglen = strlen(msg);
    bp = 0;
    for (x = 0; x < msglen; x++) {
        if (msg[x] == '\n') {
            buf[bp] = 0; 
            bbsi_csendp(3, bbsi_nextkey(), "ADDMSG", "LINE", buf);
            bp = 0;
        } else if (isprint(msg[x])) buf[bp++] = msg[x];
    }

    bbsi_csendp(3, bbsi_nextkey(), "ADDMSG", "END", "");
}

void bbs_putplanfile(char *msg) {
    char buf[255];
    int bp, x, msglen;

    bbsi_csendp(3, bbsi_nextkey(), "PUTPLANFILE", "START", "");

    msglen = strlen(msg);
    bp = 0;
    for (x = 0; x < msglen; x++) {
        if (msg[x] == '\n') {
            buf[bp] = 0; 
            bbsi_csendp(3, bbsi_nextkey(), "PUTPLANFILE", "LINE", buf);
            bp = 0;
        } else if (isprint(msg[x])) buf[bp++] = msg[x];
    }

    bbsi_csendp(3, bbsi_nextkey(), "PUTPLANFILE", "END", "");
}

int bbs_deletemsg(int baseid, int msgnum) {
    char baseid_st[10];

    strcpy(baseid_st, bbsi_itoa(baseid));
    strcpy(bbs_waitkey, bbsi_nextkey());
    bbsi_csendp(3, bbs_waitkey, "DELETEMSG", baseid_st, bbsi_itoa(msgnum));
    return bbs_handlemsg();
}

void bbs_getbasenames(void) {
    strcpy(bbs_waitkey, bbsi_nextkey());
    bbsi_csendp(1, bbs_waitkey, "GETBASES");
    bbs_handlemsg();
}

void bbs_getoptions(void) {
    strcpy(bbs_waitkey, bbsi_nextkey());
    bbsi_csendp(1, bbs_waitkey, "GETOPTS");
    bbs_handlemsg();
}

void bbs_setoptions(int scrlen, unsigned long options, char *subs) {
    char scrlen_st[10];

    strcpy(scrlen_st, bbsi_itoa(scrlen));
    strcpy(bbs_waitkey, bbsi_nextkey());
    bbsi_csendp(4, bbs_waitkey, "SETOPTS", scrlen_st, bbsi_utoa(options), subs);
    bbs_handlemsg();
}

int bbs_getlastmsg(int baseid) {
    strcpy(bbs_waitkey, bbsi_nextkey());
    bbsi_csendp(2, bbs_waitkey, "GETLASTMSG", bbsi_itoa(baseid));
    return bbs_handlemsg();
}

int bbs_setlastmsg(int baseid, int msgnum) {
    char baseid_st[10];

    strcpy(baseid_st, bbsi_itoa(baseid));
    strcpy(bbs_waitkey, bbsi_nextkey());
    bbsi_csendp(3, bbs_waitkey, "SETLASTMSG", baseid_st, bbsi_itoa(msgnum));
    return bbs_handlemsg();
}

void bbs_setpassword(char *pass) {
    bbsi_csendp(2, bbsi_nextkey(), "SETPASSWD", pass);
}

void bbs_finduser(char *partialname) {
    strcpy(bbs_waitkey, bbsi_nextkey());
    bbsi_csendp(2, bbs_waitkey, "FINDUSER", partialname);
    bbs_handlemsg();
}

int bbs_newuser(char *name, char *pass) {
    strcpy(bbs_waitkey, bbsi_nextkey());
    bbsi_csendp(3, bbs_waitkey, "NEWUSER", name, pass);
    return bbs_handlemsg();
}

int bbs_userexist(char *name) {
    strcpy(bbs_waitkey, bbsi_nextkey());
    bbsi_csendp(2, bbs_waitkey, "USEREXIST", name);
    return bbs_handlemsg();
}

void bbs_shutdown(char *mode, char *message) {
    bbsi_csendp(3, bbsi_nextkey(), "SHUTDOWN", mode, message);
}

void bbs_onfrom(char *onfrom) {
    bbsi_csendp(2, bbsi_nextkey(), "ONFROM", onfrom);
}
